/*************************************************************************/
/*   MIDI tone generator (FM synthesis) with demo song (SMAF sequencer)  */
/*   for STMicroelectronics STM32F103xx boards                           */
/*                                                                       */
/*     CQ-STARM   (CQ publishing)                                        */
/*     STBee      (Strawberry Linux)                                     */
/*     STBee Mini (Strawberry Linux)                                     */
/*                                                                       */
/*   NON COMMERCIAL USE ONLY                                             */
/*                                                                       */
/*   2010/09/19 Modified for TGSTCQ/TGSTBEE/TGSTBM                       */
/*   2008/04/19 Rev. 1.01 by pcm1723                                     */
/*                                                                       */
/*************************************************************************/

#include "tonegen.h"


#ifdef LCD
char banner_str[] = "STM32F103xx MIDI\nFM ToneGen V1.01";

const uint8_t cgram[CGRAMSIZE] = {
//  0x01, 0x01, 0x01, 0x01, 0x01, 0x0f, 0x1f, 0x0e, // quarter
//  0x00, 0x00, 0x00, 0x00, 0x1f, 0x0e, 0x0e, 0x00, // rest
  0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x1f, // C
  0x10, 0x10, 0x10, 0x17, 0x17, 0x17, 0x10, 0x1f, // C, Key ON
  0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x04, 0x04, 0x1f, // C#, D#, F#, G#, A#
  0x1f, 0x1b, 0x11, 0x1b, 0x1f, 0x04, 0x04, 0x1f, // C#, D#, F#, G#, A#, Key ON
  0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x1f, // E, B
  0x01, 0x01, 0x01, 0x1d, 0x1d, 0x1d, 0x01, 0x1f, // E, B, Key ON
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x1f, // D, F, G, A
  0x00, 0x00, 0x00, 0x0e, 0x0e, 0x0e, 0x00, 0x1f  // D, F, G, A, Key ON
};
#endif 

#ifdef SEQUENCER

/* Demo songs */
const demo_st demosong[] = {
#ifdef EXTRADEMO
  {Chopin66, 
   {2, 2, 2, 2}, // inst = Electric grand piano
   "Chopin:Fantaisie-Impromptu Op.66"},
#endif
  {arabesq1, 
   {46, 46, 46, 46}, // inst = Orchestral Harp
   "C.A.Debussy :\n\"Arabesque 1\"   "}
#ifdef EXTRADEMO
 ,{invent13,
   {4, 4, 4, 4},     // inst = E.Piano1
   "J.S.Bach:BWV.784\n\"Inventionen 13\""}
#endif
}; // const unsigned char* demosong[]
#endif

// Operator mult code to actual multiplier conversion
const int mult_tab[16] = {
// code:       0   1  2  3  4   5   6   7   8   9  10  11  12  13  14  15
// multiplier: 0.5 1  2  3  4   5   6   7   8   9  10  10  12  12  15  15
               1,  2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 20, 24, 24, 30, 30
}; // const int mult_tab[]

// LFO increment table for 4ms tick
// value = 0x10000 * tick_period * LFO_freq
const int LFO_inc_tab[4] = {
// LFO freq:  1.9Hz 4.2Hz 6.1Hz 7.2Hz
              498,  1101, 1599, 1887
              
}; // const int LFO_inc_tab[4]

int MAXMIDICH;
int MAXUNITS;
int SEQTICK1ms;
int SEQTICK4ms;
int FS_ASC;
uint seq_duration ;
int seq_tick;
int tick1ms  = 0;
int t4ms     = 0;
int8_t midithru = 0;
#ifdef SEQUENCER
int demoflag = 1;
int loopflag = 0;
int demo_no  = 0;
int demosongs = sizeof(demosong) / sizeof (demo_st);
#endif
int basech = 0;
int eos;
int panel_mode;
int last_note;
int midi_mon_part;
int sys_mode_index;
int oct_ofs;
int note_ofs;
int debounce;
uint32_t inc_sw_q, dec_sw_q;
uint16_t inc_sw_rl, dec_sw_rl;
int8_t  sw_val;
int8_t  re_state;
uint32_t panel_sw_q;
int8_t sw_val;
int demosave;
int8_t  serial_midi = 0;

volatile int t1_tick = 0;
#ifdef USEWAVEQUE
volatile int wq_wr_ix, wq_rd_ix;
short wave_que[WAVEQUESIZE];
#endif

#ifdef syn_CV
volatile int spiq_wr_ix, spiq_rd_ix;
int8_t spi_que[SPIQUESIZE];
#endif

// convert 4-bit keycode to 6dB/oct key scaling value
// KSL6.0dB = -(n - 32)/2 [dB]
// KSL6.0dB(in lb_Q6 format) = (64/6)*(n - 32)/2 = (64/12)*(n - 32)
//                           = (64/12)*(12*oct + k - 32)
//                           = 64*(oct - 3) + (16/3)*(k + 4)
const uint8_t ks60tab[12] = {
   (16*4)/3,   (16*5)/3,   (16*6)/3,   (16*7)/3,  // C ,C#,D ,D#
   (16*8)/3,   (16*9)/3,   (16*10)/3,  (16*11)/3, // E ,F ,F#,G
   (16*12)/3,  (16*13)/3,  (16*14)/3,  (16*15)/3  // G#,A ,A#,B
}; // unsigned char ks60tab[] 

unit_prop_t unit_prop[ABS_MAX_UNITS];

ch_prop_t ch_prop[MAXSEQCH];

#ifdef LCD
uint8_t lcd_buf[LCD_BUFSIZE];
uint8_t lcd_buf2[LCD_BUFSIZE];
uint8_t *lcd_buf_p;
uint8_t *lcd_buf_w;
uint8_t lcd_row, lcd_col, lcd_buf_ind;
int16_t lcd_disp_timer;
uint8_t lcd_disp_row, lcd_disp_col, lcd_cur_pos;
int16_t lcd_blink_timer;
uint8_t lcd_blink_flag;
#endif

master_prop_t master_prop;
int sysx_cnt;

uint8_t sysx_buf[MAXSYSEXLEN];

#ifdef REVERB
int     delay_mem_size;
int16_t delay_ms = DEFAULT_DELAY_ms;
int16_t delaymem[DELAYMEMSIZE];
#endif

#ifdef SEQUENCER
void demo_off(void)
{
  if (demoflag | demosave) {
    int i;
    for (i = 0; i <MAXMIDICH; i++) 
    all_note_off(&(ch_prop[i]));
//   GM_system_on();  // all note off and init 
#ifdef LCD
    lcd_buf_p = lcd_buf; // restore default screen
    lcd_draw_panel(panel_mode);
#endif
    demoflag = 0;
    demosave = 0;
  } // if (demoflag)
} // void demo_off()
#endif

// main
int main(void)
{
  volatile uint16_t count=0;
#ifndef USEWAVEQUE
  uint32_t lw;
#endif
	long int w0;
#ifdef REVERB
  long w1;
#endif
  int i;
  long int sum;
//  ch_prop_t   *c_p;
#ifdef REVERB
  int echo_dep, delayp;
#endif


#ifdef SEQUENCER
  demosave = demoflag;
  #ifdef LCD
    demoflag = 0; 
  #endif
#endif                              
#ifdef STM32F103
  stm32_port_setup();
#endif
// search for a configuration with fs >= 15 kHz
  for (  i  = MAX_CONFIG - 1;
       (15 > sys_config[fastchip][i].tick1ms) & (0 < i);
       i--) {} 
//  sys_mode_index = i; 
#ifdef DEMOONLY
  sys_mode_index = 0;
#else
  sys_mode_index = 4;
#endif
  last_note = -1;
  midi_mon_part = 0; // MIDI monitor channel
  wq_wr_ix = 0; // wave queue write index = 0
  wq_rd_ix = 0; // wave queue read  index = 0
// fill up wave queue with 0
  for (i = 0; i < WAVEQUESIZE; i++) wave_que[i] = 0;
#ifdef REVERB
  for (i = 0; i < DELAYMEMSIZE; i++) delaymem[i] = 0;
  delayp = 0;
  echo_dep = 128;
#endif
  set_config(sys_mode_index);
  LED_ON;
  hard_reset(); // hard reset
  enable_interrupt();
#ifdef LCD
  panel_mode = MIDI_MON;
  lcd_init();
  lcd_draw_panel(panel_mode);
  lcd_buf_ind = 0;
#else 
  #ifdef SEQUENCER
    init_seq();
  #endif
#endif
  w0  = 0;
  sum = 0; // clear initial wave output
  while (1){ // Main sampling rate loop
  	if (CHECKMIDI) { // serial received
     LED_OFF;        // OFF
     mididec(GETMIDI); // decode MIDI data
     LED_ON;         // ON
  } // if (CHECKMIDI) ...
  else
#ifdef USEWAVEQUE
    if(wq_wr_ix != wq_rd_ix) // wave queue has room to write
#else
    if(t1_tick)
#endif
  {
    LED_OFF;       // OFF
#ifdef REVERB
    sum += (w0 >> 1); // mix dry 50% : reverb 50 %
#endif

#ifdef syn_CV
    switch (0x03 & count) { // SPI xmit timing
      case 0:
        if (CHECK_SPIQ) {
          if (SPI_1ST_BYTE) SS_LOW;  // make SS low before 1st byte
          XMIT_SPI;
        } // if (CHECK_SPIQ)
        break;
      case 2:
        if (SPI_1ST_BYTE) SS_HIGH; // make SS high for idle
        break;
      default:
        break;
    } // switch ()
#endif

#ifdef USEWAVEQUE
    wave_que[wq_wr_ix] = sum;
    wq_wr_ix = (wq_wr_ix + 1) & (WAVEQUESIZE-1);
#else
		lw = (unsigned long)(0x8000 ^ sum) << 12; // adjust for DAC bit allocation
		DAC3DAT = lw; // output wave to DAC3
#endif
    t1_tick =0;   // clear T1 tick flag
#ifdef USEWAVEQUE
    if (1 == tick1ms) { // EG update timing
      for (i = 0; i < MAXUNITS; i++) {
        unit_prop_t *u_p = &(unit_prop[i]);
        eg(u_p, &(u_p->op_prop[0]));
        eg(u_p, &(u_p->op_prop[1]));
      } // for 
    } // if 
#else
    if      (1 == tick1ms) { // EG for unit 0
      unit_prop_t *u_p = &(unit_prop[0]);
      eg(u_p, &(u_p->op_prop[0]));
      eg(u_p, &(u_p->op_prop[1]));
    }
    else if (5 == tick1ms) { // EG for unit 1
      unit_prop_t *u_p = &(unit_prop[1]);
      eg(u_p, &(u_p->op_prop[0]));
      eg(u_p, &(u_p->op_prop[1]));
    }
    else if (9 == tick1ms) { // EG for unit 2
      unit_prop_t *u_p = &(unit_prop[2]);
      eg(u_p, &(u_p->op_prop[0]));
      eg(u_p, &(u_p->op_prop[1]));
    }
#endif
// combine output of 3 units with 1/4 scaling to avoid overflow
    sum= 0;
    for (i = 0; i < MAXUNITS; i++){
      int ssum = sum;
      sum += unit(&(unit_prop[i])) >> 2; 
//      if (2 != i) sum = ssum;
    } // for

#ifdef REVERB
    w0 = delaymem[delayp]; // read sample from delay memory
    w1 = ((w0+1) >> 1)                              // feedback of delayed sample
       + ((sum * ch_prop[0].MD_rev_send) >> 7); // mult by reverb send level of ch0
// too high send level may cause overflow, apply clipping if so
    if      ( MAXSHORT < w1) w1 =  MAXSHORT;
    else if (-MAXSHORT > w1) w1 = -MAXSHORT;
    delaymem[delayp++] = w1; // write sample in delay memory
    if (delay_mem_size <= delayp) delayp = 0; // modulo operation
#endif

    count++;
    count &= 0x3f;
    if (0 >= --tick1ms) { // 1 ms tick timing
      tick1ms = SEQTICK1ms; // restore 1ms counter
#ifdef LCD
      if (0 > --lcd_blink_timer) { // LCD blink timer
        lcd_blink_timer = LCD_BLINK_ms - 1;
        if (&(lcd_buf[0]) == lcd_buf_p) {
          lcd_blink_flag  ^= 1; // flip blink flag
        } else {                // sub screen displayed
          lcd_blink_flag   = 0; // do not blink if so
        } // else
      } // if (0 > --lcd_blink_timer) 
#endif
      t4ms++;
      t4ms &= 0x3; // divide by 4 with 1ms tick
      if (MAXUNITS > t4ms) {
        update_LFO(&(ch_prop[t4ms])); // update each channel LFO at 4ms interval
        update_porta(&(unit_prop[t4ms])); // update portament of each unit at 4ms
      } // if (MAXUNITS > t4ms)
      if (MAXUNITS > (t4ms+4)) {
        update_LFO(&(ch_prop[t4ms+4])); // update each channel LFO at 4ms interval
        update_porta(&(unit_prop[t4ms+4])); // update portament of each unit at 4ms
      } // if (MAXUNITS > t4ms)
#ifdef LCD      
      if (0 >= --debounce) { // debounce interval
        debounce = DEBOUNCE_ms; // restore debounce counter
        if ((sw_val = sw_input())) {  // INC or DEC switch released
#ifdef SEQUENCER
          if (demoflag | demosave) {     // in demo mode
            switch (abs(sw_val)) { //
              case CLICK :    // skip to next song
                seq_songsel(demo_no+sw_val);
                break;
              case PRESS :    // exit from demo mode
                demo_off();
                break;
              default: break;
            } // switch
          } // if (demoflag)
          else 
#endif
            panel_cmd(sw_val); // other commands
        } // if ((sw_val == sw_input())) ...
        else {
          sw_val = NOCHANGE;
          if (SET_CONFIG == panel_mode) lcd_cpuload();
        } // else
      } // if (0 >= debounce)
#endif      
    } // if (0 >= --tick1ms)
#ifdef LCD
    if (4 == tick1ms) {
      switch (t4ms) { 
        case 3:  // force E trigger off
          lcd_E_down();
          if (lcd_disp_timer) { // LCD display timer working
            if (0 == --lcd_disp_timer) {
              lcd_buf_p = &(lcd_buf[0]); // restore pointer to normal screen
#ifdef SEQUENCER
              if (demosave) {
                demoflag = demosave;
                demosave = 0;
                init_seq();
              } // if (demosave)
#endif
            } // if
          } // if 
          break;
#ifdef LCD_4bit
        case 2: // LCD low nibble output timing
          lcd_refresh(1);
          break;
        case 1: // force E trigger off
          lcd_E_down();
          break;
#endif
        case 0: // LCD high nibble output timing
          lcd_refresh(0);
          break;
        default: break;
      } // switch 
    } // if (4 == tick1ms)
#endif
#ifdef SEQUENCER
    if (demoflag) seq(); // SMAF sequencer top level
#endif
    LED_ON;       // 
    }  // if 
  } // while (1) 
} // int main()
