/******************************************************/
/*                                                    */
/* sysxdec.c : MIDI System Exclusive message decoder  */
/*                                                    */
/* NON COMMERCIAL USE ONLY                            */
/*                                                    */
/* 2007/06/26 Rev. 1.01 by pcm1723                    */
/*                                                    */
/******************************************************/

#include "tonegen.h"

int  sx_ix;
int32_t sx_mfr_id;
int  sx_dev_id, sx_sub_id1, sx_sub_id2, sx_csum;

/* get a byte from Sys Ex buffer */
int sx_getb(void)
{
  uint8_t b;
  b = sysx_buf[sx_ix++]; // get a byte from Sys Ex buffer
  sx_csum += b;           // update checksum
  return (b);
} // int sx_getb()

/* get 2-nibbles to form a byte from Sys_Ex buffer */
int sx_get2h(void)
{
  uint8_t b, h;
  h = sysx_buf[sx_ix++]; // get a byte from Sys_Ex buffer
  sx_csum += h;           // update checksum
  h <<= 4;                // move to high nibble
  b = sysx_buf[sx_ix++]; // get a byte from Sys_Ex buffer
  sx_csum += b;           // update checksum
  h += (0x0f & b);        // combine high and low nibble
  return (h);
} // int sx_get2h()

/* Scale/Octave tuning 1-byte form (-64cent .. +63cent) */
void sx_so_tuning1(void)
{
  uint i, j, ch_bit, bitmap;
  int ofs;
  ch_bit = 0;
  for (i = 0; i < 3; i++) ch_bit = (ch_bit << 7) + sx_getb(); // build ch bitmap
  for (i = 0; i < 12; i++) { // 12 semitones (C to B) scan
    ofs = centab[sx_getb()]; // convert cent to note fraction (-41 .. +40)
    bitmap = ch_bit; // remember tuning enabled channel
    for (j = 0; j < MAXMIDICH; j++) { // channel bitmap test
      if (0x01 & bitmap) ch_prop[j].so_tune_ofs[i] = ofs; // tuning enabled
      bitmap >>= 1; // try next channel
    } // for (j = 0; ...
  } // for (i = 0; ...
} // void sx_so_tuning()

/* Non-realtime universal system exclusive */
void sx_non_rt_univ(void) 
{
  switch (sx_sub_id1) { // decode by sub-ID 1
    case 0x08 : // MIDI Tuning standard
      if (0x08 == sx_sub_id2) { // scale/octave tuning 1-byte
        sx_so_tuning1();
      } // if (0x08 == ...
      break;
    case 0x09 : // General MIDI 
      if (0x01 <= sx_sub_id2) {
#ifdef SEQUENCER
        demo_off();
#endif
        GM_system_on(); // GM1 and GM2 ON
#ifdef LCD
        if (MIDI_MON == panel_mode) { //
          lcd_buf[MON_NN_DISP_OFS  ] = 'G';
          lcd_buf[MON_NN_DISP_OFS+1] = (1 < sx_sub_id2 ? '2' : 'M');
        }
#endif
      }
      break;
    default:
      break;
  } // switch (sx_sub_id1)
} // void sx_non_rt_univ()

/* Realtime universal system exclusive */
void sx_rt_univ() 
{
  switch (sx_sub_id1) { // decode by sub-ID 1
    case 0x04 : // Device control
      switch (sx_sub_id2) { // sub ID
        case 0x01 :  // MASTER VOLUME
          sx_getb(); // discard LSB of volume
          master_prop.MD_master_volume = sx_getb(); // get MSB of volume
          break;
        case 0x03 :  // MASTER FINE TUNING
          sx_getb(); // discard LSB of fine tuning
          master_prop.MD_master_fine_tune = sx_getb(); // get MSB of fine tuning
          break;
        case 0x04 :  // MASTER COARSE TUNING
          sx_getb(); // discard LSB of fine tuning
          master_prop.MD_master_coarse_tune = sx_getb(); // get MSB of coarse tuning
          break;
       default: break;
      } // switch (sx_sub_id2) ...
      break;
    case 0x08 : // MIDI Tuning standard
      if (0x08 == sx_sub_id2) { // scale/octave tuning 1-byte
        sx_so_tuning1();
      } // if (0x08 == ...
      break;
    default:
      break;
  } // switch (sx_sub_id1)
} // void sx_rt_univ()

void Roland_Sys_Ex()
{
   int i;
   uint32_t adrs;
   if ((0x10 == sx_dev_id) & (0x42 == sx_sub_id1) & (0x12 == sx_sub_id2)) { //
     adrs = 0;
     for (i = 0; i < 3; i++) adrs = (adrs << 8) + sx_getb(); // get address
     if (0x40007f == adrs) { // GS reset
       if (0x00 == sx_getb()) {
         sx_getb(); // calculate checksum
         if (0x00 == (0x7f & sx_csum)) {
#ifdef SEQUENCER
           demo_off();
#endif
           GM_system_on();
#ifdef LCD
           if (MIDI_MON == panel_mode) { //
             lcd_buf[MON_NN_DISP_OFS  ] = 'G';
             lcd_buf[MON_NN_DISP_OFS+1] = 'S';
           }
#endif
         } // if (0x00 == (0x7f
       } // if (0x00 == sx_getb())
     } // if (0x40007f == adrs) ...
   } // if  
} // void Roland_Sys_Ex()

void XG_param_change(uint32_t adrs)
{
  int i;
  char c;
  switch (adrs) { // decode
    case 0x00007e : // XG reset
      if (0x00 == sx_getb()) {
#ifdef SEQUENCER
        demo_off();
#endif
        GM_system_on();
#ifdef LCD
        if (MIDI_MON == panel_mode) { //
          lcd_buf[MON_NN_DISP_OFS  ] = 'X';
          lcd_buf[MON_NN_DISP_OFS+1] = 'G';
        }
#endif
      }
      break;
#ifdef LCD
    case 0x060000 : // character display
        lcd_disp_timer = LCD_DISP_TICKS; // set display timer
        lcd_clear(lcd_buf2);
        lcd_buf_p = &(lcd_buf2[0]);
        i = 0;
        c = sx_getb(); // get first char
        while (0xf7 != c) { // exit at EOX
          if (LCD_BUFSIZE > i)  
            lcd_buf2[i++] = ((' ' > c) ? ' ' : c);
          c = sx_getb(); // get next char
        } // while
#endif      
    default: break;
  } // switch (adrs) ...
} // void XG_param_change

void MA2_sysx(int id2, int ch, int type)
{
  int i;
  switch (id2) {
    case 0x08 : // parameter set 
      if (MAXMIDICH > ch) { // supported part
        ch_prop_t *c_p = &(ch_prop[ch]);
//        struct ma2vdat *ma2v = (void *)&(c_p->vdat); // current voice data buffer
        ma2vdat *ma2v = (void *)&(c_p->vdat); // current voice data buffer
        switch (type) { // decode
          case 0x00 : // voice data
            for (i =  0; i <  4; i++) (*ma2v)[i] = sx_get2h(); // bank/inst/LFO/FB/ALG
            sx_get2h(); // skip one byte
            for (i =  4; i <  9; i++) (*ma2v)[i] = sx_get2h(); // copy OP0 data
            for (i = 10; i < 15; i++) (*ma2v)[i] = sx_get2h(); // copy OP1 data
            decode_voice(c_p);
//            copy_voice(c_p);
            break;
          default : break;
        } // switch (i)
      } // if 
      break;
    default: break;
  } // switch
} // void MA2_sysx()

void YAMAHA_Sys_Ex()
{
  int i;
  uint32_t adrs;
  adrs = sx_sub_id2;
  for (i = 0; i < 2; i++) adrs = (adrs << 8) + sx_getb(); // get address
  switch (sx_sub_id1) { //  decode by model ID
    case 0x4c : // XG (generic) parameter change
      if (0x10 == (0xf0 & sx_dev_id)) {
        XG_param_change(adrs);
      } //
      break;
    case 0x03 : // MA-2
        MA2_sysx(sx_sub_id2, 0xff & (adrs >> 8), 0xff & adrs);
      break;
    default: break;
  } // switch (sx_sub_id1) ...
} // void YAMAHA_Sys_Ex()

void sysx_dec(int len)
{
  int i;
  if ((0xf0 == sysx_buf[0]) & (0xf7 == sysx_buf[len-1])) { // SOX and EOX checck
#ifdef LCD
    if (SHOW_Sys_Ex == panel_mode) { // show exclusive message
      int i;
      lcd_clear(lcd_buf);
      for (i = 0; (i < sysx_cnt) & (i < 16); i++) {
        lcd_puthex2(sysx_buf[i]);
      } // for (i
    } // if (SHOE_Sys_Ex == panel_mode)
#endif
    sx_ix = 1; // start from Manufacture ID
    sx_mfr_id = sx_getb(); // assume 1-byte ID
    if (0 == sx_mfr_id) {  // 3-byte ID
      for (i = 0; i < 2; i++) {
        sx_mfr_id = (sx_mfr_id << 8) + sx_getb();
      } // for
    } // if (0 ==
    sx_dev_id  = sx_getb(); // Device ID
    sx_sub_id1 = sx_getb(); // Sub ID 1 / Model ID
    sx_sub_id2 = sx_getb(); // Sub ID 2 / Command ID
    sx_csum    = 0;         // clear checksum
    switch (sx_mfr_id) {    // decode by Mfr ID
      case 0x7e :  // Non-realtime universal system exclusive message
        sx_non_rt_univ();
        break;
      case 0x7f :  // Realtime universal system exclusive message
        sx_rt_univ();
        break;
      case 0x41 :  // Roland
        Roland_Sys_Ex();
        break;
      case 0x43 :  // YAMAHA
        YAMAHA_Sys_Ex();
        break;
      default:
        break;
    } // switch (mfr_id)
  } // if ... 
} // void sysxdec()
